/*

*/

import renderer.scene.*;
import renderer.scene.util.*;
import renderer.models_L.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
   This file defines a hierarchical scene out of a single instance of Model_1.
<p>
   The graph for this scene is shown below. Since two Position objects
   share a reference to the same Model_1 object, this graph is not a tree,
   it is a "directed acyclic graph" (DAG). In general, scene graphs can be
   DAGs since multiple Position objects can share a Model object, and
   Position objects can also share a (nested) Position object.
<p>
<pre>{@code
           Scene
          /     \
         /       \
    Camera     List<Position>
                     |
                     |
                  Position
                 /   |    \
                /    |     \
          Matrix   Model     List<Position>
            R     (empty)   /              \
                           /                \
                          /                  \
                   Position             Position
                    /      \            /      /
                   /        \          /      /
               Matrix        \      Matrix   /
                 I            \      TSR    /
                               \           /
                                \         /
                                  Model_1
}</pre>
*/
public class ShowModel_2a
{
   public static void main(String[] args)
   {
      // Create the Scene object that we shall render.
      final Scene scene = new Scene("ShowModel_2a");

      // Create the top level Position.
      final Position p = new Position("top");

      // Add the top level Position to the Scene.
      scene.addPosition(p);

      // Add two nested Positions to the top level Position.
      final Position p1 = new Position("left");
      final Position p2 = new Position("right");
      p.addNestedPosition(p1, p2);

      // Add a single instance of Model_1 to the Scene.
      final Model m1 = new Model_1();
      ModelShading.setColor(m1, Color.red);
      // Add a reference to Model m1 to each of Positions p1 and p2.
      p1.setModel(m1);
      p2.setModel(m1);

      // Initialize the nested matrices in the Positions.
      p2.getMatrix().mult( Matrix.translate(1, -2-Math.sqrt(2), 0) )
                    .mult( Matrix.scale(0.5, 0.5, 1) )
                    .mult( Matrix.rotateZ(-45) );

      DrawSceneGraph.draw(scene, "SG_ShowModel_2a");
      DrawSceneGraphTopologicalSort.draw(scene, "SG_ShowModel_2a_TS");


      // Create a FrameBuffer to render our Scene into.
      final int vp_width  = 1024;
      final int vp_height = 1024;
      final FrameBuffer fb = new FrameBuffer(vp_width, vp_height);

      //PipelineLogger.debug = true;

      for (int i = 0; i <= 36; ++i)
      {
         // Push the models away from where the camera is.
         p.matrix2Identity()
          .mult( Matrix.translate(0, 0, -5) )
          .mult( Matrix.rotateZ(10*i) );

         // Render again.
         fb.clearFB(Color.white);
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_ShowModel_2a_Frame%02d.ppm", i));
      }
   }
}
