/*
 * Renderer 8. The MIT License (MIT).
 * Copyright (c) Roger Kraft (rlkraft@pnw.edu).
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.scene.util.ModelShading;
import renderer.scene.util.DrawSceneGraph;
import renderer.models_L.Sphere;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;
import javax.swing.JFrame;
import java.awt.BorderLayout;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import java.awt.event.ComponentListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.Timer;

/**
   Draw an animation of a solar system with a sun, planet, and moon.
<p>
   In this version, the orbit of the planet is independent of the
   rotation of the sun, and the orbit of the moon is independent
   of the rotation of the planet.
<pre>{@code
           Scene
             |
             |
          Position
         /   |     \
        /    |      \
  Matrix   Model    nested Positions
    I     (empty)    /           \
                    /             \
               Position            Position
               /     \             /   |   \
              /       \           /    |    \
         Matrix     Model     Matrix  Model  nested Positions
           R        (sun)       RT   (empty)  /            \
                                             /              \
                                            /                \
                                      Position            Position
                                      /     \              /     \
                                     /       \            /       \
                                 Matrix     Model      Matrix     Model
                                   R       (planet)     RTR      (moon)
}</pre>
*/
public class SolarSystem_v2a implements ActionListener, KeyListener, ComponentListener
{
   private int fps;
   private Timer timer = null;

   private double planetOrbitRadius = 5.0;
   private double   moonOrbitRadius = 1.0;

   private double planetOrbitRot = 0.0;
   private double   moonOrbitRot = 0.0;

   private double    sunAxisRot = 0.0;
   private double planetAxisRot = 0.0;
   private double   moonAxisRot = 0.0;

   private double ecliptic = 7.0; // angle of the ecliptic plane

   private final Scene scene;
   private final Position solarSys_p;
   private final Position sun_p;
   private final Position planetMoon_p;
   private final Position planet_p;
   private final Position moon_p;

   protected boolean takeScreenshot = false;
   protected int screenshotNumber = 0;

   private final JFrame jf;
   private final FrameBufferPanel fbp;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public SolarSystem_v2a()
   {
      /*
         See the above picture of the tree that this code creates.
      */

      // Create the Scene object that we shall render
      scene = new Scene("Solar System 2a");

      // Create the Position that holds the whole solar system.
      scene.addPosition(new Position("SolarSystem"));
      solarSys_p = scene.getPosition(0);

      // Create the sun.
      solarSys_p.addNestedPosition(new Position(new Sphere(1.0, 10, 10), "Sun"));
      ModelShading.setColor(solarSys_p.getNestedPosition(0).getModel(), Color.yellow);
      sun_p = solarSys_p.getNestedPosition(0);

      // Create the Position that holds the planet-moon system.
      solarSys_p.addNestedPosition(new Position("PlanetMoon"));
      planetMoon_p = solarSys_p.getNestedPosition(1);

      // Create the planet.
      planetMoon_p.addNestedPosition(new Position(new Sphere(0.5, 10, 10), "Planet"));
      ModelShading.setColor(planetMoon_p.getNestedPosition(0).getModel(), Color.blue);
      planet_p = planetMoon_p.getNestedPosition(0);

      // Create the moon.
      planetMoon_p.addNestedPosition(new Position(new Sphere(0.2, 10, 10), "Moon"));
      ModelShading.setColor(planetMoon_p.getNestedPosition(1).getModel(), Color.green);
      moon_p = planetMoon_p.getNestedPosition(1);

      DrawSceneGraph.draw(scene, "SG_SolarSystem_v2a");


      // Create a FrameBufferPanel that will hold a FrameBuffer.
      final int width  = 1024;
      final int height = 1024;
      fbp = new FrameBufferPanel(width, height);
      fbp.getFrameBuffer().setBackgroundColorFB(Color.black);
      fbp.getFrameBuffer().getViewport().setBackgroundColorVP(Color.black);

      // Create a JFrame that will hold the FrameBufferPanel.
      jf = new JFrame("Renderer 9");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      jf.pack();
      jf.setLocationRelativeTo(null);
      jf.setVisible(true);

      // Create event handler objects for events from the JFrame.
      jf.addKeyListener(this);
      jf.addComponentListener(this);

      fps = 20;
      timer = new Timer(1000/fps, this); // ActionListener
      timer.start();

      print_help_message();
   }


   // Implement the ActionListener interface.
   @Override public void actionPerformed(ActionEvent e)
   {
      //System.out.println( e );

      // Push the solar system away from where the camera is.
      solarSys_p.matrix2Identity()
                .mult( Matrix.translate(0, 0, -8) );
      // Rotate the plane of the ecliptic
      // (rotate the solar system's xz-plane about the x-axis).
      solarSys_p.getMatrix().mult( Matrix.rotateX(ecliptic) );

      // Set the model matrices for the nested positions.
      sun_p.matrix2Identity()
           .mult( Matrix.rotateY(sunAxisRot) );

      planetMoon_p.matrix2Identity()
                  .mult( Matrix.rotateY(planetOrbitRot) )
                  .mult( Matrix.translate(planetOrbitRadius, 0, 0) );

      planet_p.matrix2Identity()
              .mult( Matrix.rotateY(planetAxisRot) );

      moon_p.matrix2Identity()
            .mult( Matrix.rotateY(moonOrbitRot) )
            .mult( Matrix.translate(moonOrbitRadius, 0, 0) )
            .mult( Matrix.rotateY(moonAxisRot) );

      // Update the parameters for the next frame.
      sunAxisRot -= 10.0;
      planetOrbitRot += 1.0;
      planetAxisRot -= 5.0;
      moonOrbitRot += 5.0;
      moonAxisRot += 10.0;

      // Render again.
      final FrameBuffer fb = this.fbp.getFrameBuffer();
      fb.clearFB();
      Pipeline.render(scene, fb);
      if (takeScreenshot)
      {
         fb.dumpFB2File(String.format("Screenshot%03d.png", screenshotNumber),
                        "png");
         ++screenshotNumber;
         takeScreenshot = false;
      }
      fbp.update();
   }


   // Implement the KeyListener interface.
   @Override public void keyPressed(KeyEvent e){}
   @Override public void keyReleased(KeyEvent e){}
   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      final char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('d' == c && e.isAltDown())
      {
         System.out.println();
         System.out.println( scene );
      }
      else if ('d' == c)
      {
         scene.debug = ! scene.debug;
         Clip.debug = scene.debug;
      }
      else if ('D' == c)
      {
         Rasterize.debug = ! Rasterize.debug;
      }
      else if ('a' == c)
      {
         Rasterize.doAntiAliasing = ! Rasterize.doAntiAliasing;
         System.out.print("Anti-aliasing is turned ");
         System.out.println(Rasterize.doAntiAliasing ? "On" : "Off");
      }
      else if ('g' == c)
      {
         Rasterize.doGamma = ! Rasterize.doGamma;
         System.out.print("Gamma correction is turned ");
         System.out.println(Rasterize.doGamma ? "On" : "Off");
      }
      else if ('f' == c)
      {
         fps -= 1;
         if (0 > fps) fps = 0;
         System.out.println("fps = " + fps);
         setFPS();
      }
      else if ('F' == c)
      {
         fps += 1;
         System.out.println("fps = " + fps);
         setFPS();
      }
      else if ('s' == c)
      {
         // Stop the animation.
         timer.stop();
      }
      else if ('S' == c)
      {
         // Start the animation.
         timer.start();
      }
      else if ('p' == c)
      {
         scene.getCamera().perspective = ! scene.getCamera().perspective;
         final String p = scene.getCamera().perspective ? "perspective"
                                                        : "orthographic";
         System.out.println("Using " + p + " projection");
      }
      else if ('e' == c)
      {
         ecliptic -= 1;
         System.out.println("ecliptic = " + ecliptic);
      }
      else if ('E' == c)
      {
         ecliptic += 1;
         System.out.println("ecliptic = " + ecliptic);
      }
      else if ('+' == c)
      {
         takeScreenshot = true;
      }

      // Set up the camera's view volume.
      if (scene.getCamera().perspective)
      {
         scene.getCamera().projPerspective(-1, 1, -1, 1, 1);
      }
      else
      {
         scene.getCamera().projOrtho(-6, 6, -6, 6, -1);
      }
   }


   public void setFPS()
   {
      timer.stop();
      if (fps > 0)
      {
         timer = new Timer(1000/fps, this);
         timer.start();
      }
   }


   // Implement the ComponentListener interface.
   @Override public void componentMoved(ComponentEvent e){}
   @Override public void componentHidden(ComponentEvent e){}
   @Override public void componentShown(ComponentEvent e){}
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );
      /*
      System.out.printf("JFrame [w = %d, h = %d]: " +
                        "FrameBufferPanel [w = %d, h = %d].\n",
                        jf.getWidth(), jf.getHeight(),
                        fbp.getWidth(), fbp.getHeight());
      */
      // Get the new size of the FrameBufferPanel.
      final int w = fbp.getWidth();
      final int h = fbp.getHeight();

      // Create a new FrameBuffer that fits the FrameBufferPanel.
      final Color bg1 = fbp.getFrameBuffer().getBackgroundColorFB();
      final Color bg2 = fbp.getFrameBuffer().getViewport()
                                            .getBackgroundColorVP();
      final FrameBuffer fb = new FrameBuffer(w, h, bg1);
      fb.vp.setBackgroundColorVP(bg2);
      fbp.setFrameBuffer(fb);
      // allow the timer event to update the new framebuffer and window
   }


   private void print_help_message()
   {
      System.out.println("Use the 'd' key to toggle debugging information on and off.");
      System.out.println("Use the 'Alt-d' key combination to print the Scene data structure.");
      System.out.println("Use the 'a' key to toggle antialiasing on and off.");
      System.out.println("Use the 'g' key to toggle gamma correction on and off.");
      System.out.println("Use the f/F keys to slow down or speed up the frame rate.");
      System.out.println("Use the 's/S' key to stop/Start the animation.");
      System.out.println("Use the 'p' key to toggle between parallel and orthographic projection.");
      System.out.println("Use the e/E keys to change the angle of the ecliptic plane.");
      System.out.println("Use the '+' key to save a \"screenshot\" of the framebuffer.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new SolarSystem_v2a()
      );
   }
}
