/*
 * Renderer 9. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

package renderer.pipeline;

import renderer.scene.*;
import renderer.scene.util.CheckModel;
import renderer.framebuffer.*;
import static renderer.pipeline.PipelineLogger.*;

import java.awt.Color;

/**
   This renderer takes as its input a {@link Scene} data structure
   and a {@link FrameBuffer.Viewport} within a {@link FrameBuffer}
   data structure. This renderer mutates the {@link FrameBuffer.Viewport}
   so that it is filled in with the rendered image of the geometric
   scene represented by the {@link Scene} object.
<p>
   This implements our eigth rendering pipeline. This renderer
   implements "hierarchical scenes" by recursively traversing the
   DAG of nested {@link Position}s below each of a {@link Scene}'s
   {@link Position} objects. As the renderer traverses deeper into
   the DAG of nested {@link Position}s, it accumulates a "current
   transformation matrix" that transforms each {@link Vertex} from a
   {@link Model}'s local coordinate system to the {@link Camera}'s
   (shared) view coordinate system. The recursive traversal of the
   scene's DAG is not a new pipeline stage, so there are still six
   pipeline stages.
*/
public final class Pipeline
{
   // Mostly for compatibility with renderers 1 through 3.
   public static Color DEFAULT_COLOR = Color.white;

   /**
      Mutate the {@link FrameBuffer}'s default {@link FrameBuffer.Viewport}
      so that it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   public static void render(final Scene scene, final FrameBuffer fb)
   {
      render(scene, fb.vp); // render into the default viewport
   }


   /**
      Mutate the {@link FrameBuffer}'s given {@link FrameBuffer.Viewport}
      so that it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param vp     {@link FrameBuffer.Viewport} to hold rendered image of the {@link Scene}
   */
   public static void render(final Scene scene, final FrameBuffer.Viewport vp)
   {
      PipelineLogger.debugScene = scene.debug;

      logMessage("\n== Begin Rendering of Scene (Pipeline 1): " + scene.name + " ==");

      // For every Position in the Scene, render the Position's Model.
      for (final Position position : scene.positionList)
      {
         PipelineLogger.debugPosition = position.debug;

         if ( position.visible )
         {
            // Begin a pre-order, depth-first-traversal from this Position.
            render_position(scene, position, Matrix.identity(), vp);
         }
         else
         {
            logMessage("==== Hidden position: " + position.getName());
         }
      }
      logMessage("== End Rendering of Scene (Pipeline 1) ==");
   }


   /**
      Recursively renderer a {@link Position}.
      <p>
      This method does a pre-order, depth-first-traversal of the tree of
      {@link Position}'s rooted at the parameter {@code position}.
      <p>
      The pre-order "visit node" operation in this traversal first updates the
      "current transformation matrix", ({@code ctm}), using the {@link Matrix}
      in {@code position} and then renders the {@link Model} in {@code position}
      using the updated {@code ctm} in the {@link Model2View} stage.

      @param scene     the {@link Scene} that we are rendering
      @param position  the current {@link Position} object to recursively render
      @param ctm       current model-to-view transformation {@link Matrix}
      @param vp       {@link FrameBuffer.Viewport} to hold rendered image of the {@link Scene}
   */
   private static void render_position(final Scene scene,
                                       final Position position,
                                       final Matrix ctm,
                                       final FrameBuffer.Viewport vp)
   {
      // Opening message.
      if ( position.getModel() != null )
      {
         logMessage("==== Render position: " + position.getName() + " ====");
      }
      else
      {
         logMessage("==== Render position (no model) ====");
      }

      logMessage("---- Transformation matrix:\n" + position.getMatrix());

      // Update the current model-to-view transformation matrix.
      final Matrix ctm2 = ctm.times( position.getMatrix() );

      // Render this Position's Model, if the model exits and is visible.
      if ( position.getModel() != null && position.getModel().visible )
      {
         logMessage("====== Render model: "
                            + position.getModel().getName() + " ======");

         CheckModel.check(position.getModel());

         // Mostly for compatibility with renderers 1 through 3.
         if (  position.getModel().colorList.isEmpty()
           && !position.getModel().vertexList.isEmpty())
         {
            for (int i = 0; i < position.getModel().vertexList.size(); ++i)
            {
               position.getModel().addColor( DEFAULT_COLOR );
            }
            System.err.println("***WARNING: Added default color to model: "
                              + position.getModel().getName() + ".");
         }

         logVertexList("0. Model       ", position.getModel());

         // 1. Apply the current model-to-view coordinate transformation.
         final Model model1 = Model2View.model2view(position, ctm2);

         logVertexList("1. View        ", model1);

         // 2. Apply the Camera's normalizing view-to-camera coordinate transformation.
         final Model model2 = View2Camera.view2camera(model1, scene.getCamera());

         logVertexList("2. Camera      ", model2);
         logColorList("2. Camera      ", model2);
         logPrimitiveList("2. Camera       ", model2);

         // 3. Clip primitives to the camera's near plane.
         final Model model3 = NearClip.clip(model2, scene.getCamera());

         logVertexList("3. Near_Clipped", model3);
         logColorList("3. Near_Clipped", model3);
         logPrimitiveList("3. Near_Clipped", model3);

         // 4. Apply the Camera's projection transformation.
         final Model model4 = Projection.project(model3, scene.getCamera());

         logVertexList("4. Projected   ", model4);

         // 5. Clip primitives to the camera's view rectangle.
         final Model model5 = Clip.clip(model4);

         logVertexList("5. Clipped     ", model5);
         logColorList("5. Clipped     ", model5);
         logPrimitiveList("5. Clipped     ", model5);

         // 6. Rasterize every visible primitive into pixels.
         Rasterize.rasterize(model5, vp);

         logMessage("====== End model: "
                            + position.getModel().getName() + " ======");
      }
      else
      {
         if ( position.getModel() != null )
         {
            logMessage("====== Hidden model: "
                               + position.getModel().getName() + " ======");
         }
         else
         {
            logMessage("====== Missing model ======");
         }
      }

      // Recursively render every nested Position of this Position.
      for (final Position p : position.nestedPositions)
      {
         if ( p.visible )
         {
            // Do a pre-order, depth-first-traversal from this nested Position.
            render_position(scene, p, ctm2, vp); // recursion
         }
         else
         {
            logMessage("====== Hidden position "
                               + position.getName() + " ====");
         }
      }

      // Closing message.
      if ( position.getModel() != null )
      {
         logMessage("==== End position: " + position.getName() + " ====");
      }
      else
      {
         logMessage("==== End position (no model) ====");
      }
   }



   // Private default constructor to enforce noninstantiable class.
   // See Item 4 in "Effective Java", 3rd Ed, Joshua Bloch.
   private Pipeline() {
      throw new AssertionError();
   }
}
