/*
 * Renderer 9. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

package renderer.scene.util;

import renderer.scene.*;

import java.io.File;
import java.io.PrintWriter;
import java.lang.Runtime;

/**
   This program converts a {@link Scene} data structure into
   a DOT description of the scene. The DOT description is
   then processed by the dot.exe program to produce a
   graphical image of the scene data structure.
<p>
   Create a <code>scene.png</code> file from a <code>scene.dot</code>
   file with the following command line.
   <pre>{@code
   > dot.exe -Tpng -O scene.dot
   }</pre>
<p>
   See
<br><a href="https://www.graphviz.org/Documentation.php" target="_top">
             https://www.graphviz.org/Documentation.php</a>
*/
public class DrawSceneGraph_save
{
   private static int positionNumber;
   private static int modelNumber;

   /**
      This method converts the {@link Scene} data structure into a dot
      language description, then writes the dot code into a dot file,
      then runs the dot.exe program on the dot file to create a png
      image of the scene data structure.

      @param scene     {@link Scene} that needs to be converted to a dot description
      @param fileName  base name for the dot and png files
   */
   public static void draw(final Scene scene, final String fileName)
   {
      // Convert the scene data structure into a dot language description.
      final String dotDescription = scene2dot(scene);

      // Write the dot language description stored in dotDescription
      // into a dot file. Then use the dot.exe program to convert the
      // dot file into a png file.
      try
      {
         // Create the (empty) dot file.
         String baseName = fileName;
         java.io.PrintWriter out = new PrintWriter(
                                      new File(baseName + ".dot"));

         // Write the dot commands into the dot file.
         out.print( dotDescription );
         out.close();

         // Create a command-line for running the dot.exe program.
         String[] cmd = {"C:\\graphviz-2.38\\release\\bin\\dot.exe",
                         "-Tpng",
                         baseName + ".dot",
                         "-o",
                         baseName + ".png"};

         File dot = new File("C:\\graphviz-2.38\\release\\bin\\dot.exe");
         if(dot.exists() && !dot.isDirectory())
         {
            // Execute the command-line to create the png file.
            Runtime.getRuntime().exec(cmd);
         }
         else
         {
            System.out.println("Please consider installing GraphViz:");
            System.out.println("  https://graphviz.org/download/");
            System.out.println("Or using Graphviz Visual Editor:");
            System.out.println("  http://magjac.com/graphviz-visual-editor/");
         }
      }
      catch (Exception e)
      {
         System.out.println( e );
      }
   }


   /**
      This method generates a dot language description of the
      DAG rooted at {@code scene}.
      <p>
      This method generates the dot code for the forest of top-level
      positions just below the scene node. Each position node just
      below the scene node is the root of a DAG. This method calls
      the <code>position2dot()</code> method to recursively traverse
      the DAG of each top-level position. Every position has attached
      to it a {@link Model}, so every position node has attached to it
      a model node.

      @param  scene {@link Scene} that needs to be converted to a dot description
      @return a {@link String} containing the dot language description of the scene
   */
   public static String scene2dot(final Scene scene)
   {
      String result = "digraph {\n";

      // https://graphviz.org/docs/attrs/ordering/
      result += "graph [ordering=\"out\"];\n";

      // Scene node.
      result += "scene  [label=\"Scene\"];\n";

      positionNumber = -1;
      modelNumber = -1;

      // Create a node and an edge for each top-level Position and its Model.
      for (int i = 0; i < scene.positionList.size(); ++i)
      {
         ++positionNumber;

         // Position node name.
         final String pNodeName = "_p" + positionNumber;

         // Position node and label.
         result += pNodeName + " ";
         result += "[label=\"Position: " + scene.getPosition(i).getName() + "\"];\n";

         // Position edge.
         result += "scene -> " + pNodeName + ";\n";

         // Model node name and label.
         ++modelNumber;
         final Model modelReference = scene.getPosition(i).getModel();
         final String mNodeName;
         // Check that a model object exists.
         if (null != modelReference)
         {
            // Check if the Model is being reused.
            final int modelAddress = computeModelAddress(scene, modelReference);
            // Model node name.
            mNodeName = "_m" + modelAddress;
            // Model node and label (if the Model hasn't already been used).
            if (modelAddress == modelNumber)
            {
               result += mNodeName + " ";
               result += "[label=\"Model: " + modelReference.getName() + "\"];\n";
            }
         }
         else
         {
            // Model node name.
            mNodeName = "_m" + modelNumber;
            // Model node and label.
            result += mNodeName + " ";
            result += "[label=\"Model: empty\"];\n";
         }

         // Model edge.
         result += pNodeName + " -> " + mNodeName + ";\n";

         // Convert this position's nested positions into a dot description.
         if ( ! scene.getPosition(i).nestedPositions.isEmpty() )
         {
            // positionList node.
            result += pNodeName + "_List ";
            result += "[label=\"List<Position>\"];\n";
            // positionList edge.
            result += pNodeName + " -> " + pNodeName + "_List;\n";
            // Nested positions.
            result += position2dot(scene, scene.getPosition(i), pNodeName + "_List");
         }
      }

      result += "}\n";

      return result;
   }


   /**
      This <code>position2dot()</code> method is essentially a
      preorder traversal of the DAG of {@link Position} nodes.
      <p>
      This method generates a dot language description of the
      DAG rooted at the give {@code position}.
      <p>
      {@code nodeLabel} is the id that has been assigned to the
      node for the given {@code position}. Every position has
      attached to it a {@link Model}, so every position node
      has attached to it a model node.

      @param position  {@link Position} that needs to be converted to a dot description
      @param nodeName  the {@link String} name that has been assigned to {@code position}
      @return a {@link String} containing the dot language description of the position
   */
   public static String position2dot(final Scene scene,
                                     final Position position,
                                     final String nodeName)
   {
      String result = "";

      // Create nodes and edges for each nested position.
      for (int i = 0; i < position.nestedPositions.size(); ++i)
      {
         ++positionNumber;

         // Position node name.
         final String newNodeName = "_p" + positionNumber;

         // Position node and label.
         result += newNodeName + " ";
         result += " [label=\"Position\"];\n";

         // Position edge.
         result += nodeName + " -> " + newNodeName + ";\n";

         // Model node name and label.
         ++modelNumber;
         final Model modelReference = position.getNestedPosition(i).getModel();
         final String mNodeName;
         // Check that a model object exists.
         if (null != modelReference)
         {
            // Check if the Model is being reused.
            final int modelAddress = computeModelAddress(scene, modelReference);
            // Model node name.
            mNodeName = "_m" + modelAddress;
            // Model node and label (if the Model hasn't already been used).
            if (modelAddress == modelNumber)
            {
               result += mNodeName + " ";
               result += "[label=\"Model: " + modelReference.getName() + "\"];\n";
            }
         }
         else
         {
            // Model node name.
            mNodeName = "_m" + modelNumber;
            // Model node and label.
            result += mNodeName + " ";
            result += "[label=\"Model: empty\"];\n";
         }

         // Model edge.
         result += newNodeName + " -> " + mNodeName + ";\n";

         // Recursively convert this position's nested positions into a dot description.
         if ( ! position.getNestedPosition(i).nestedPositions.isEmpty() )
         {
            // positionList node.
            result += newNodeName + "_List ";
            result += " [label=\"List<Position>\"];\n";
            // positionList edge.
            result += newNodeName + " -> " + newNodeName + "_List;\n";
            // Nested positions.
            result += position2dot(scene, position.getNestedPosition(i), newNodeName + "_List");
         }
      }

      return result;
   }


   private static int computeModelAddress;
   private static int computeModelAddress(Scene scene, Model model)
   {
      computeModelAddress = -1;
      for (Position p : scene.positionList)
      {
         final int rv = computeModelAddressHelper(p, model);
         if (rv >= 0) return rv;
      }

      throw new IllegalArgumentException("model not found in the scene");
   }

   private static int computeModelAddressHelper(Position p, Model model)
   {
      ++computeModelAddress;
      if ( model == p.getModel() )
      {
         return computeModelAddress;
      }
      for (Position p2 : p.nestedPositions)
      {
         final int rv = computeModelAddressHelper(p2, model);
         if (rv >= 0) return rv;
      }
      return -1;
   }


   // Private default constructor to enforce noninstantiable class.
   // See Item 4 in "Effective Java", 3rd Ed, Joshua Bloch.
   private DrawSceneGraph_save()
   {
      throw new AssertionError();
   }
}
